pub mod general;
pub mod pokemon;

use std::{
    fmt::Display,
    ops::{Index, IndexMut},
    sync::atomic::{AtomicU32, Ordering},
};

use general::{Gender, Generation, Language, Stat};

pub trait RNG {
    fn get(&self) -> u32;
}

pub fn prng(prev: u32) -> u32 {
    (0x41c64e6d * prev).wrapping_add(0x6073)
}

pub fn aprng(prev: u32) -> u32 {
    (0x6c078965 * prev).wrapping_add(0x1)
}

/// the Pseudo-Random Number Generator used in pokemon.
///
/// used for:
/// - encrypting pokemon data in the save file (see http://projectpokemon.org/wiki/Pokemon_NDS_Structure?#Encryption)
/// - generating the personality value in wild pokemon
/// - generating the IVs in wild pokemon
/// - determining the species of wild pokemon
/// - determining the lottery number
pub struct PRng(AtomicU32);

impl RNG for PRng {
    fn get(&self) -> u32 {
        self.0
            .fetch_update(Ordering::Relaxed, Ordering::Relaxed, |v| {
                Some((0x41c64e6d * v).wrapping_add(0x6073))
            })
            .expect("closure never returns none")
    }
}

/// the Alternative Pseudo-Random Number Generator
///
/// "but what about second rng?"
///
/// see https://projectpokemon.org/docs/other/prng-in-pokémon-r38/
pub struct APRng(AtomicU32);

impl RNG for APRng {
    fn get(&self) -> u32 {
        self.0
            .fetch_update(Ordering::Relaxed, Ordering::Relaxed, |v| {
                Some((0x6c078965 * v).wrapping_add(0x1))
            })
            .expect("closure never returns none")
    }
}

pub struct IVs {
    pub hp: u8,
    pub atk: u8,
    pub def: u8,
    pub spa: u8,
    pub spd: u8,
    pub spe: u8,
}

impl Index<Stat> for IVs {
    type Output = u8;

    fn index(&self, index: Stat) -> &Self::Output {
        use Stat::*;

        match index {
            HitPoints => &self.hp,
            Attack => &self.atk,
            Defense => &self.def,
            SpAtk => &self.spa,
            SpDef => &self.spd,
            Speed => &self.spe,
        }
    }
}

impl IndexMut<Stat> for IVs {
    fn index_mut(&mut self, index: Stat) -> &mut Self::Output {
        use Stat::*;

        match index {
            HitPoints => &mut self.hp,
            Attack => &mut self.atk,
            Defense => &mut self.def,
            SpAtk => &mut self.spa,
            SpDef => &mut self.spd,
            Speed => &mut self.spe,
        }
    }
}

pub struct EVs {
    pub hp: u8,
    pub atk: u8,
    pub def: u8,
    pub spa: u8,
    pub spd: u8,
    pub spe: u8,
}

impl Index<Stat> for EVs {
    type Output = u8;

    fn index(&self, index: Stat) -> &Self::Output {
        use Stat::*;

        match index {
            HitPoints => &self.hp,
            Attack => &self.atk,
            Defense => &self.def,
            SpAtk => &self.spa,
            SpDef => &self.spd,
            Speed => &self.spe,
        }
    }
}

impl IndexMut<Stat> for EVs {
    fn index_mut(&mut self, index: Stat) -> &mut Self::Output {
        use Stat::*;

        match index {
            HitPoints => &mut self.hp,
            Attack => &mut self.atk,
            Defense => &mut self.def,
            SpAtk => &mut self.spa,
            SpDef => &mut self.spd,
            Speed => &mut self.spe,
        }
    }
}

pub struct CurrentMoveSet(LearnedMove, LearnedMove, LearnedMove, LearnedMove);

pub struct LearnedMove {
    inner: Move,
    pp_up: u8,
}

pub enum Move {}
