use std::io::Read;

use crate::general::{
    Ability, BinaryGender, EVs, Gender, IVs, Language, Move, MoveSet, Nature, PokeBall, Species,
    StatSpread, StatusCondition, Trainer,
};

pub mod gen5;

pub trait StringConverter {
    fn to_string(bytes: impl Read) -> String;
    fn to_bytes(string: String) -> Box<dyn Read>;
}

pub struct Pokemon<P> {
    pkm: P,
    pending_progressions: Vec<Progression>,
}

pub trait PKM {
    fn ot(&self) -> &Trainer;

    fn gain_xp(&mut self, xp: u32) -> Vec<Progression>;
    fn defeated(&mut self, other: Species);
    fn use_item(&mut self, item: Item) -> Result<(), ItemUseError>;
    fn damage(&mut self, hp: u16);
    fn apply_status(&mut self, status: StatusCondition);

    fn species(&self) -> Species;
    fn ivs(&self) -> IVs;
    fn evs(&self) -> EVs;
    fn is_shiny(&self) -> bool;
    fn nickname(&self) -> Option<String>;
    fn held_item(&self) -> Option<Item>;

    fn current_hp(&self) -> u16;
    fn max_hp(&self) -> u16;
    fn effective_base_stats(&self) -> StatSpread<u16>;
    fn move_set(&self) -> MoveSet;

    fn display_name(&self) -> String {
        self.nickname()
            .unwrap_or_else(|| format!("{:?}", self.species())) // todo: multi lang
    }
    fn fainted(&self) -> bool {
        self.current_hp() == 0
    }
    fn health(&self) -> f32 {
        self.current_hp() as f32 / self.max_hp() as f32
    }
}

pub enum Progression {
    LevelUp,
    Evolution(Species),
    LearnMove(Move),
}

impl<P> Pokemon<P>
where
    P: PKM,
{
    pub fn into_inner(self) -> P {
        self.pkm
    }
}

impl<P> From<P> for Pokemon<P> {
    fn from(value: P) -> Self {
        Self {
            pkm: value,
            pending_progressions: Vec::new(),
        }
    }
}
